#!/usr/bin/python

###
# Copyright (c) 2009 Jason Deiman jay@splitstreams.com
#
# This software in it's source or binary forms is available for use under
# the GNU GPL Version 2.  See http://www.gnu.org/licenses/gpl-2.0.txt for the
# full details.
###

__version__ = '0.3.4'
__svnversion__ = '$Id: clcheck.py 306 2010-05-03 01:28:22Z jay $'

"""
This script is set up to read a config file and query craigslist for search
terms and store the results.  When changed, it will send an email alert
with the new items.
"""

import sys , os , re , ConfigParser , optparse , urllib , urllib2
import time , smtplib , email.utils , datetime , platform
from email.mime.text import MIMEText

DEBUG = False

if DEBUG:
    import traceback

confDefaults = {
        'clCat': 'for sale' ,
        'alertTo': '' ,
        'minPrice': 'min' ,
        'maxPrice': 'max' ,
        'smtpServer': '' ,
        'smtpPort': '25' ,
        'smtpUser': '' ,
        'smtpPass': '' ,
        'smtpSSL': 'false' ,
        'smtpUseTLS': 'false' ,
}

try:
    import sqlite3
except Exception , e:
    print >> sys.stderr , 'There is no module "sqlite3" that should be part ' \
            'of the standard library.  Please download and install it'
    sys.exit(1)

class ConfigError(Exception):
    pass

class SearchError(Exception):
    pass

class ResultError(Exception):
    pass

class SmtpError(Exception):
    pass

class DBError(Exception):
    pass

class SearchResult(object):
    detExtractRe = re.compile(r'<hr>.+?Date: ([^<]+)<br>.+?Reply to:\s+' + \
            r'(?:<a href="mailto:[^"]+">([^<]+)</a>|(see below))' + \
            r'.*?<div id="userbody">(.*?)(?:<!--\s+START CLTAGS -->.*?)?' + \
            r'<br><br><ul[^>]*>.+?<li>(?:.*?Location:\s*([^<]+)<)?.*?' + \
            r'PostingID:\s+([^<]+)<' , re.I | re.M | re.S)
    imgExtractRe = re.compile( r'<img src="(http://images.craigslist' + \
            r'.org/[^"]+)"[^>]+>+' ,
            re.I | re.M | re.S)
    dNormRe = re.compile(r'^(.*?),\s+(\d+)(:\d+[AP]M).*$' , re.I)
    def __init__(self , clBase , href , desc):
        self.clBase = clBase
        href = self._trimHref(href)
        self.href = href
        self.fullUrl = 'http://%s%s' % (clBase , href)
        self.desc = desc
        self.emailTo = ''
        self.datePosted = None
        self.details = ''
        self.location = ''
        self.images = []
        self.postId = -1

        self._getDetails()

    def _stripItems(self , items):
        items = list(items)
        if items[0] == None:
            del items[0]
        elif items[1] == None:
            del items[1]
        elif items[2] == None:
            del items[2]
        else:
            raise ResultError , 'Unexpected results: %s' % items
        ret = []
        for i in items:
            if i:
                i = i.strip()
            ret.append(i)
        return ret

    def _getDetails(self):
        r = urllib2.urlopen(self.fullUrl)
        if r.code < 200 or r.code >= 300:
            raise ResultError , 'An error occured getting the page at ' \
                    '%s: %s : %s' % (self.fullUrl , r.code , r.msg)
        page = r.read()
        m = self.detExtractRe.search(page)
        if not m:
            raise ResultError , 'Unparsable page at %s' % self.fullUrl

        clGroups = self._stripItems(m.groups())
        self.emailTo = clGroups[1]
        self.datePosted = datetime.datetime.strptime(
                self._normalizeDate(clGroups[0]) , '%Y-%m-%d, %I:%M%p')
        self.details = clGroups[2]
        self.location = ('' , clGroups[3])[clGroups[3] == None]
        try:
            self.postId = int(clGroups[4])
        except Exception , e:
            raise ResultError , 'Invalid postid: %s' % clGroups[4]
        # Get Images
        self.images = self.imgExtractRe.findall(page)

    def _normalizeDate(self , dstr):
        # Break up the date time
        m = self.dNormRe.match(dstr)
        if not m:
            raise ResultError , 'The date string is not properly formatted:' \
                  '%s' % dstr
        date = m.group(1)
        hour = m.group(2)
        remaining = m.group(3)
        # Pad the hour if necessary
        if len(hour) == 1:
            hour = '0%s' % hour
        return '%s, %s%s' % (date , hour , remaining)

    def _trimHref(self , href):
        """
        This is to strip an href that is absolute instead of relative.
        """
        if href.startswith('http'):
            repl = 'http://%s' % self.clBase
            return href.replace(repl , '')
        return href

    def getTsTime(self):
        return time.mktime(self.datePosted.timetuple())

class SearchItem(object):
    clCats = {
            # Community
            'community': 'ccc' , 
            'activity partners': 'act' ,
            'artists': 'ats' ,
            'childcare': 'kid' ,
            'community-general': 'com' ,
            'groups': 'grp' ,
            'local news and views': 'vnn' ,
            'lost & found': 'laf' ,
            'musicians': 'muc' ,
            'pets': 'pet' ,
            'politics': 'pol' ,
            'rideshare': 'rid' ,
            'volunteers': 'vol' ,
            # Events
            'events': 'eee' , 
            'classes': 'cls' ,
            'events-events': 'eve' ,
            # Gigs
            'gigs': 'ggg' , 
            'adult gigs': 'adg' ,
            'computer gigs': 'cpg' ,
            'creative gigs': 'crg' ,
            'crew gigs': 'cwg' ,
            'domestic gigs': 'dmg' ,
            'event gigs': 'evg' ,
            'labor gigs': 'lbg' ,
            'talent gigs': 'tlg' ,
            'writing gigs': 'wrg' ,
            # Housing
            'housing': 'hhh' ,
            'all housing wanted': 'hsw' ,
            'apts wanted': 'hou' ,
            'apts/housing for rent': 'apa' ,
            'housing swap': 'swp' ,
            'office & commercial': 'off' ,
            'parking & storage': 'prk' ,
            'real estate - all': 'rea' ,
            'real estate - by broker': 'reb' ,
            'real estate - by owner': 'reo' ,
            'real estate wanted': 'rew' ,
            'rooms & shares': 'roo' ,
            'rooms wanted': 'sha' ,
            'sublet/temp wanted': 'sbw' ,
            'sublets & temporary': 'sub' ,
            'vacation rentals': 'vac' ,
            # Jobs
            'jobs': 'jjj' ,
            'admin/office jobs': 'ofc' ,
            'business jobs': 'bus' ,
            'customer service jobs': 'csr' ,
            'education jobs': 'edu' ,
            'engineering jobs': 'egr' ,
            'etcetera jobs': 'etc' ,
            'finance jobs': 'acc' ,
            'food/bev/hosp jobs': 'fbh' ,
            'general labor jobs': 'lab' ,
            'government jobs': 'gov' ,
            'healthcare jobs': 'hea' ,
            'human resource jobs': 'hum' ,
            'internet engineering jobs': 'eng' ,
            'legal jobs': 'lgl' ,
            'manufacturing jobs': 'mnu' ,
            'marketing jobs': 'mar' ,
            'media jobs': 'med' ,
            'nonprofit jobs': 'npo' ,
            'real estate jobs': 'rej' ,
            'retail/food/hospitality jobs': 'rfh' ,
            'retail/wholesale jobs': 'ret' ,
            'sales jobs': 'sls' ,
            'salon/spa/fitness': 'spa' ,
            'science jobs': 'sci' ,
            'security jobs': 'sec' ,
            'skilled trades jobs': 'trd' ,
            'software jobs': 'sof' ,
            'systems/networking jobs': 'sad' ,
            'tech support jobs': 'tch' ,
            'transport jobs': 'trp' ,
            'tv video radio jobs': 'tfr' ,
            'web design jobs': 'web' ,
            'writing jobs': 'wri' ,
            # Personals
            'personals': 'ppp' , 
            'casual encounters': 'cas' ,
            'men seeking men': 'm4m' ,
            'men seeking women': 'm4w' ,
            'misc romance': 'msr' ,
            'missed connections': 'mis' ,
            'rants & raves': 'rnr' ,
            'strictly platonic': 'stp' ,
            'women seeking men': 'w4m' ,
            'women seeking women': 'w4w' ,
            # Resumes
            'resumes': 'res' , 
            # For Sale
            'for sale': 'sss' ,
            'art & crafts': 'art' ,
            'auto parts': 'pts' ,
            'baby & kid stuff': 'bab' ,
            'barter': 'bar' ,
            'bicycles': 'bik' ,
            'boats': 'boa' ,
            'books': 'bks' ,
            'business': 'bfs' ,
            'cars & trucks - all': 'cta' ,
            'cars & trucks - by dealer': 'ctd' ,
            'cars & trucks - by owner': 'cto' ,
            'cds / dvds / vhs': 'emd' ,
            'clothing': 'clo' ,
            'collectibles': 'clt' ,
            'computers & tech': 'sys' ,
            'electronics': 'ele' ,
            'farm & garden': 'grd' ,
            'free stuff': 'zip' ,
            'furniture - all': 'fua' ,
            'furniture - by dealer': 'fud' ,
            'furniture - by owner': 'fuo' ,
            'games & toys': 'tag' ,
            'garage sales': 'gms' ,
            'fs-general': 'for' ,
            'household': 'hsh' ,
            'items wanted': 'wan' ,
            'jewelry': 'jwl' ,
            'materials': 'mat' ,
            'motorcycles/scooters': 'mcy' ,
            'musical instruments': 'msg' ,
            'photo/video': 'pho' ,
            'recreational vehicles': 'rvs' ,
            'sporting goods': 'spo' ,
            'tickets': 'tix' ,
            'tools': 'tls' ,
            # Services
            'services': 'bbb' ,
            'adult services': 'ads' ,
            'automotive services': 'aos' ,
            'beauty services': 'bts' ,
            'computer services': 'cps' ,
            'creative services': 'crs' ,
            'event services': 'evs' ,
            'financial services': 'fns' ,
            'household services': 'hss' ,
            'labor & moving': 'lbs' ,
            'legal services': 'lgs' ,
            'lessons & tutoring': 'lss' ,
            'real estate services': 'rts' ,
            'skilled trade services': 'sks' ,
            'small biz ads': 'biz' ,
            'therapeutic services': 'thp' ,
            'travel/vacation': 'trv' ,
            'write/edit/trans': 'wet' ,
    }
    dbTableSQL = """
        CREATE TABLE previous (
            sid INTEGER PRIMARY KEY ,
            ts REAL ,
            postid INTEGER ,
            heading TEXT ,
            url TEXT
        );
        CREATE INDEX ts_prev_idx ON previous (ts);
        CREATE INDEX heading_prev_idx ON previous (heading);
        CREATE INDEX postid_prev_idx ON previous (postid);
    """
    dExtractRe = re.compile(r'<!--\s+sphinx.*?(<p>.*?)<br>' , re.I | re.M |
            re.S)
    resExtractRe = re.compile(r'<p>[^<]+<a href="([^"]+)">([^<]+)</a>.*?</p>' ,
            re.I)
    br2nlRe = re.compile(r'<br\s*/?>' , re.I | re.M | re.S)
    reduceNlRe = re.compile(r'\n{3,}' , re.M | re.S)
    remHtmlRe = re.compile(r'</?\w+[^>]+>' , re.M | re.S)

    def __init__(self , conf , sectName):
        self.sectName = sectName
        self.search = conf.get(sectName , 'search')
        self.clBase = conf.get(sectName , 'clBase')
        clCat = conf.get(sectName , 'clCat')
        if not clCat in self.clCats.keys():
            raise ConfigError , 'Invalid category name (clCat): %s' % clCat
        self.clCat = clCat
        self.dbLoc = conf.get(sectName , 'dbLoc')
        self.minPrice = conf.get(sectName , 'minPrice')
        self.maxPrice = conf.get(sectName , 'maxPrice')
        self.alertTo = conf.get(sectName , 'alertTo')
        self.smtpServer = conf.get(sectName , 'smtpServer')
        self.smtpPort = conf.getint(sectName , 'smtpPort')
        self.smtpUser = conf.get(sectName , 'smtpUser')
        self.smtpPass = conf.get(sectName , 'smtpPass')
        self.smtpSSL = conf.getboolean(sectName , 'smtpSSL')
        self.smtpUseTLS = conf.getboolean(sectName , 'smtpUseTLS')
        if self.smtpSSL and self.smtpUseTLS:
            raise ConfigError , 'You cannont speficy BOTH SSL and TLS'
        self.newItems = []
        # db connection
        self._db = None
        self._initDb = False
        self._errored = False
        try:
            self._connectDb()
        except sqlite3.OperationalError , e:
            raise DBError , '''An error occured connecting to the database, 
            %s, no searches will be performed for "%s"''' % (self.dbLoc , 
                    sectName)
            self._errored = True

    def _connectDb(self):
        if not os.path.isfile(self.dbLoc):
            self._initDb = True
            if not os.path.isdir(os.path.dirname(self.dbLoc)):
                os.makedirs(os.path.dirname(self.dbLoc))
        self._db = sqlite3.connect(self.dbLoc)
        self._db.text_factory = str
        if self._initDb:
            c = self._db.cursor()
            c.executescript(self.dbTableSQL)
            c.close()

    def _getUrl(self):
        qstring = urllib.urlencode({'query': self.search , 
                'minAsk': self.minPrice ,
                'maxAsk': self.maxPrice
        })
        url = 'http://%s/search/%s?%s' % (self.clBase , 
                self.clCats[self.clCat] , qstring)
        return url

    def _getResults(self , r):
        global DEBUG
        results = []
        if r.code < 200 or r.code >= 300:
            if DEBUG:
                print >> sys.stderr , r.headers.headers
            raise SearchError , 'Error: %s %s' % (r.code , r.msg)
        page = r.read()
        # First extract data portion
        m = self.dExtractRe.search(page)
        if not m:
            if DEBUG:
                print page
            raise SearchError , 'Unrecognized search results for %s' % \
                    self.sectName
        data = m.group(1)
        del page
        del m
        items = self.resExtractRe.findall(data)
        c = self._db.cursor()
        for href , desc in items:
            q = 'SELECT sid FROM previous WHERE heading = ?'
            c.execute(q , (desc,))
            res = c.fetchone()
            if res and res[0]:
                # We already have this one
                continue
            try:
                r = SearchResult(self.clBase , href , desc)
            except Exception , e:
                print >> sys.stderr , 'RESULT ERROR: %s' % e
                continue
            results.append(r)
        return results

    def _processResults(self , results):
        c = self._db.cursor()
        for r in results:
            q = '''INSERT INTO previous (ts , postid , heading , url)
                   VALUES (? , ? , ? , ?)
                '''
            c.execute(q , (r.getTsTime() , r.postId , r.desc , r.fullUrl))
            self._db.commit()
            self.newItems.append(r)
        c.close()
        self._db.close()

    def _stripHtml(self , h):
        # Change breaks to nl
        h = self.br2nlRe.sub('\n' , h)
        h = self.reduceNlRe.sub('\n\n' , h)
        h = self.remHtmlRe.sub('' , h)
        return h

    def _getMail(self):
        body = ''
        div = ('=' * 72) + '\n'
        for r in self.newItems:
            body += div
            body += '''Heading: %s
Location: %s
Date: %s
URL: %s
Details: %s
                    ''' % (r.desc , r.location , r.datePosted.ctime() ,
                           r.fullUrl , self._stripHtml(r.details))
            if len(r.images) > 0:
                body += '\nImages:\n'
                for i in r.images:
                    body += '\t%s\n' % i
            if body[-1] != '\n':
                body += '\n'
            body += div + '\n'
        m = MIMEText(body)
        m['To'] = self.alertTo
        m['From'] = self.alertTo
        m['Subject'] = 'CLWatch: %s' % self.sectName
        m['Date'] = email.utils.formatdate(localtime=True)
        m['Message-ID'] = email.utils.make_msgid()
        return m

    def _checkSmtpResp(self , r):
        return int(r[0]) >=200 and int(r[0]) < 300

    def searchWeb(self):
        global DEBUG
        if DEBUG:
            print >> sys.stderr , 'starting search'
        if self._errored:
            return False
        url = self._getUrl()
        resp = urllib2.urlopen(url)
        if DEBUG:
            print >> sys.stderr , 'Getting url: %s' % url
        if DEBUG:
            print >> sys.stderr , 'Request completed'
        try:
            results = self._getResults(resp)
        except SearchError , e:
            print >> sys.stderr , 'An error occured for your search, %s: %s' \
                    % (self.sectName , e)
            return False
        if DEBUG:
            print >> sys.stderr , 'Got results'
        self._processResults(results)
        return len(self.newItems)

    def sendEmail(self):
        if self._initDb or self._errored:
            return False
        mess = self._getMail()
        if not self.smtpServer:
            # Print to stdout and exit
            print mess.get_payload()
            return False
        if len(self.newItems) < 1:
            return False
        smtp = None
        if self.smtpSSL:
            if DEBUG:
                print >> sys.stderr , 'Using ssl'
            smtp = smtplib.SMTP_SSL
        else:
            smtp = smtplib.SMTP
        s = smtp()
        if DEBUG:
            print 'Connecting to %s:%d' % (self.smtpServer , self.smtpPort)
        r = s.connect(self.smtpServer , self.smtpPort)
        if not self._checkSmtpResp(r):
            raise SmtpError , 'Email Error: %s %s' % (r[0] , r[1])
        # There may be an issue here in terms of timing...
        time.sleep(2)
        r = s.ehlo()
        if not self._checkSmtpResp(r):
            raise SmtpError , 'Email Error: %s %s' % (r[0] , r[1])
        if self.smtpUseTLS:
            if DEBUG:
                print >> sys.stderr , 'Using TLS'
            r = s.starttls()
            if not self._checkSmtpResp(r):
                raise SmtpError , 'Email Error: %s %s' % (r[0] , r[1])
        if self.smtpUser and self.smtpPass:
            r = s.login(self.smtpUser , self.smtpPass)
            if not self._checkSmtpResp(r):
                raise SmtpError , 'Email Error: %s %s' % (r[0] , r[1])
        s.sendmail(self.alertTo , [self.alertTo] , mess.as_string())
        s.quit()
        return True

def getOpts():
    usage = 'Usage %prog [options]'
    p = optparse.OptionParser(usage=usage)
    p.add_option('-c' , '--config' , dest='config' , metavar='FILE' ,
            default='' , 
            help='The path to the config file [default: %default]')
    p.add_option('-d' , '--debug' , dest='debug' , action='store_true' ,
            default=False , 
            help='Turn on debugging output [default: %default]')
    p.add_option('-V' , '--version' , dest='version' , action='store_true' ,
            default=False , help='Show version and exit')
    opts , junk = p.parse_args()
    return opts

def getConf(opts):
    global confDefaults
    conf = ConfigParser.ConfigParser(confDefaults)
    conffiles = ['./clwatch.cfg' , '/etc/clwatch.cfg' , 
            '/usr/local/etc/clwatch.cfg' , 
            os.path.expanduser('~/.clwatch.cfg')  , opts.config]
    l = conf.read(conffiles)
        
    if len(l) > 0:
        return conf
    else:
        raise ConfigError , 'Could not find a config file in: %s' % \
                conffiles

def check4470(conf , s):
    ssl = conf.getboolean(s , 'smtpSSL')
    if ssl:
        vers = [int(i) for i in platform.python_version_tuple()]
        if (vers[0] == 2 and vers[1] == 6) or (vers[0] == 3 and vers[1] == 0):
            print >> sys.stderr , 'You cannot use smtpSSL with python ' \
                    'version 2.6 or 3.0 as it is broken:\n' \
                    'http://bugs.python.org/issue4470\n' \
                    'Use TLS instead, if you can.'
            sys.exit(10)

def main():
    global DEBUG
    opts = getOpts()
    if opts.version:
        print 'Version: %s' % __version__
        sys.exit(0)
    if opts.debug:
        DEBUG = True
    try:
        conf = getConf(opts)
    except ConfigError , e:
        if DEBUG:
            traceback.print_exc()
        print >> sys.stderr , e
        print >> sys.stderr , 'Try specifying a config file with the -c option'
        sys.exit(2)
    except ConfigParser.MissingSectionHeaderError , e:
        if DEBUG:
            traceback.print_exc()
        print >> sys.stderr , 'Missing Section: %s' % e
        sys.exit(3)
    except ConfigParser.NoOptionError , e:
        if DEBUG:
            traceback.print_exc()
        print >> sys.stderr , 'No Option Error: %s' % e
        sys.exit(4)


    sections = conf.sections()
    if len(sections) < 1:
        print >> sys.stderr , 'You must specify at least one search in your ' \
                'config file'
        sys.exit(5)
    for s in sections:
        # Check to see if SSL is enabled, and if so check version number in
        # regards to bug 4470
        check4470(conf , s)
        si = SearchItem(conf , s)
        try:
            si.searchWeb()
        except Exception , e:
            if DEBUG:
                traceback.print_exc()
            print >> sys.stderr , 'Error in section %s: %s' % (s , e)
            continue
        try:
            si.sendEmail()
        except Exception , e:
            if DEBUG:
                traceback.print_exc()
            print >> sys.stderr , 'Error in section %s: %s' % (s , e)

if __name__ == '__main__':
    main()
